/*									*/
/************************************************************************/
/* linux device driver for PI PCI stage controller			*
 * (c) Matt Clark / GPL'd license					*/


/* the card suppots 3 memory regions					*
 * They are
 * name/function			size
 * 1) HPI CSR memory / used for interrupts			4k	*
 * 2) Control space memory / unused as far as I can tell	32k	*
 * 3) General purpose bus used to communicate with the motion		*
 *    controller chipset					256b	*/

/* The information I am working from is exceedingly unclear		*
 * but it appears that IO requests are blocked until pci_config 	*
 * _register(GPBINDATA) has bit 0x08 raised.  This is mapped to an	*
 * interrupt INTGPIO3 for interrupt driven IO.				*
 * The card also raises GPINT interrupts when a motor condition is 	*
 * raised.	- matt in ignorance 2003				*/

/* all comms to the board follow this structure CMD,DATA,DATA,...	*/
/* all comms goto the GP bus which has the following structure		*
 * offset	function						*
 *  0 		data I/O						*
 *  2		CMD (I only)						*/


/* this driver will have NO text interface but will support /procfs	*
 * for debugging and informational purposes				*/
#include "pi_stage.h"
// increment the version string to a current value
//#define VERSION         		"v0.41 March 2003" 
//note: no change to the kernel driver since 0.41, this just
//makes it consistent with the library - matt
#define VERSION         		"v0.62 August 2003" 
#define DEBUG_PAUSE 			250

/* allow certain variables to be overwritten by the load process	*/
static int pi_major=0;
MODULE_PARM(pi_major,"i");

static struct proc_dir_entry *proc_file; 

/* our sleep queue for waiting for io driven interrupts */
//DECLARE_WAIT_QUEUE_HEAD(pi_stage_sleep);
/* spin lock to prevent async tasklet from colliding with IO*/
spinlock_t pi_lock=SPIN_LOCK_UNLOCKED;


unsigned long pi_prof1,pi_prof2,pi_prof3,pi_prof4;

/*	driver function prototypes					*/

static ssize_t		pi_read(struct file *, char *, size_t, loff_t *);
static int 		pi_ioctl(struct inode *, struct file *, unsigned int , unsigned long); 
static int 		pi_open(struct inode *, struct file *);
static int 		pi_release(struct inode *, struct file *);
static void 		pi_int_handler(int, void *,struct pt_regs *);
int			pi_async(int, struct file *, int);
/*
inline static void 	pi_enable_io_int(void);
inline static void 	pi_disable_io_int(void);
inline static void 	pi_enable_axis_int(void);
inline static void 	pi_disable_axis_int(void);
*/
inline static void 	pi_write_control_w(int,u16);
inline static void 	pi_write_control(int,u32);
inline u32 		pi_read_control(int);
inline u16 		pi_read_control_w(int);
int 			pi_test_io_ok(void);
inline void 		pi_send_cmd(u16);
inline void 		pi_send_data(u16);
inline u16 		pi_read_data(void);
static void 		pi_stage_enable_card(void);
inline void 		pi_busy_wait_io(void);
void			pi_reset_board(void);
void 			pi_getasyncevent(unsigned long);

#define PI_DATA int

DECLARE_TASKLET(pi_tasklet,pi_getasyncevent,0);

/* utilities								*/
static void pi_ms_sleep(int);
static int pi_proc_read(char *, char **, off_t, int, int *, void *);

/* fops defines the standard interface for char devices			*
 * this driver will use write to access memory in the first instance	*
 * ioctl to perform most "user" functions				*
 * and later on I will implement mmap for raw IO			*
 * I will support 2 write modes- direct write and "scrambling" write	*
 * which will change the byte order appropriately 			*
 * I will restrict writing to row-at-a-time and include a "set row"	*
 * ioctl (which I can't see being used					*/
 
struct file_operations pi_fops = {
	read:		pi_read,
	ioctl:		pi_ioctl,
	open:		pi_open,
	release:	pi_release,
	fasync:		pi_async,
	};

int pi_async(int fd, struct file *fi, int mode){
	return fasync_helper(fd, fi, mode, &(pi_stage.async_queue));
	}

int init_module(void){
int	err,res;
char 	int_n;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
/* they shouldn't compile in any event- matt oct 2003			*/
	printk("PI: kernels earlier than 2.4.0 are not supported\n");
	return -ENODEV;
#endif

        printk("\nPI PCI stage controller %s (c)mc\n\n",VERSION); 
	if(!(pi_stage.dev=pci_find_device(PI_VENDOR_ID,PI_DEVICE_ID,pi_stage.dev))){
		printk("PI: no pi_stage card found in this machine\n");
		return -ENODEV;
		}
	/* this should never happen- trapping it for coding errors only */
	if(pi_stage.dev==NULL){
		printk("PI: card found but dev==NULL\n");
		return -ENODEV;
		}
	
	/* announce the pi_stage					*/
	printk("PI: stage contoller card found- probing for additional info\n");
/* wake the card up							*/
	if((err=pci_enable_device(pi_stage.dev))!=0){
		/* we are in deep trouble here if this fails- matt	*/
		printk("PI: init: pci_enable failed with %d exiting\n",err);
		return -ENODEV;
		}
/* At this point we <should> have a unique id for the card, however PI	*
 * don't have a unique vendor/device number pair at this time (Jan 2003)*
 * so we may have to install additional card probing here if required	*/


	/* get	the interrupt line #					*/
	if((err=pci_read_config_byte(pi_stage.dev,PCI_INTERRUPT_LINE,&int_n))!=0){
		/* There is no interrupt for this card!			*/
		printk("PI: init: no interrupt line found for card- exiting\n");
		return -ENODEV;
		}
	pi_stage.interrupt=int_n;
	
	printk("PI: card config gives interrupt line as %d\n",pi_stage.interrupt);

	pi_stage.interrupt=pi_stage.dev->irq;
	printk("PI: using interrupt line %d from pci_dev\n",pi_stage.interrupt);

	/* We claim the interrupt at open() */
	printk("PI: init_modules: int not claimed - it is claimed in open()\n");

	/* map the resources to virual memory				*/
	if(check_mem_region(pci_resource_start(pi_stage.dev,PI_STAGE_HPICSR),PI_STAGE_HPICSR_SIZE)){
		printk("PI: fatal error: PI_STAGE_HPICSR memory region is already occupied\n");
		err=-EBUSY;
		goto init_hpicsr_memory_error;
		}

	if(!request_mem_region(pci_resource_start(pi_stage.dev,PI_STAGE_HPICSR),PI_STAGE_HPICSR_SIZE,"pi_stage_BRIDGE")){
		printk("PI: PI_STAGE_HPICSR memory region can't be registered\n");
		err=-EBUSY;
		goto init_hpicsr_memory_error;
		}
	
	pi_stage.control=ioremap(pci_resource_start(pi_stage.dev,PI_STAGE_HPICSR)&PCI_BASE_ADDRESS_MEM_MASK,PI_STAGE_HPICSR_SIZE);
	
	if(pi_stage.control==NULL){
		printk("PI: fatal error: Unable to map PCI bridge chip space  (region %d, %x bytes) for device\n",PI_STAGE_HPICSR,PI_STAGE_HPICSR_SIZE);
		err=-ENODEV;
		goto init_no_control;
		}
	printk("PI: ioremap'd control space\n");

	if(check_mem_region(pci_resource_start(pi_stage.dev,PI_STAGE_GPBUS),PI_STAGE_GPBUS_SIZE)){
		printk("PI: fatal error: PI_STAGE_GPBUS memory region is already occupied\n");
		err=-EBUSY;
		goto init_gpbus_memory_error;
		}
	if(!request_mem_region(pci_resource_start(pi_stage.dev,PI_STAGE_GPBUS),PI_STAGE_GPBUS_SIZE,"pi_stage_GPBUS")){
		printk("PI: fatal error: PI_STAGE_GPBUS memory region is can't be registered\n");
		err=-EBUSY;
		goto init_gpbus_memory_error;
		}

	pi_stage.gpbus=ioremap(pci_resource_start(pi_stage.dev,PI_STAGE_GPBUS)&PCI_BASE_ADDRESS_MEM_MASK,PI_STAGE_GPBUS_SIZE);

	if(pi_stage.gpbus==NULL){
		printk("PI: fatal error: Unable to map gpbus  (region %d, %x bytes) for device\n",PI_STAGE_GPBUS,PI_STAGE_GPBUS_SIZE);
		err=-ENODEV;
		goto init_no_gpbus;
		}


	/* perform additional probing here - first enable the board	*/
	/* additional probing should be performed to id the board exclusively	*/

	// reset the board
	pi_reset_board();
	/* All probing to be over by here 				*/
	// probe the version number and get the number of axis installed
	{
	u32 version1,version2;
		pi_send_cmd(PI_GET_VERSION);
		version1=pi_read_data();
		version2=pi_read_data();
		pi_stage.naxis=(version1 & 0x00f0)>>4;
		printk("PI: Board probed, version %x, %x\n",version1,version2);
		if(((version1&0xf000)>>12)!=2){
			printk("PI: Version error,\n");
		       	printk("this card does not seem to be a PI motion controller board\n");
			printk("Contact PI, ask them to get a unique PCI id\n");
			goto init_no_gpbus;
			}
		printk("PI: Board has %d axis capability\n",pi_stage.naxis);
		}

		
	/* init global structures 1st 					*/
	/* now register the char device 				*/
	res=register_chrdev(pi_major,PI_MOD_NAME,&pi_fops);
	if (res<0){
		printk("PI: unable to register char dev using major=%d\n",pi_major);
		pi_ms_sleep(DEBUG_PAUSE);
		err=res;
		goto init_no_char_dev;
		}
	pi_major=res;
	
	printk("PI: init_module: device major=%d\n",pi_major); 
	pi_ms_sleep(DEBUG_PAUSE);
	printk("PI: init_module: installing /proc \n"); 
	pi_ms_sleep(DEBUG_PAUSE);
	proc_file=create_proc_read_entry(PI_MOD_NAME,0,NULL,pi_proc_read,NULL);
	if(!proc_file){
		printk("PI: init: error installing /proc file- sorry\n");
		pi_ms_sleep(DEBUG_PAUSE);
		}
	printk("PI: installation complete\n");
	pi_ms_sleep(DEBUG_PAUSE);
	
	//allocate ring buffer
	pi_stage.events=kmalloc(sizeof(struct pi_stage_event_t)*PI_N_EVENTS,GFP_KERNEL);
	if(pi_stage.events==NULL){
		printk("PI: Error: Cannot allocate ring buffer memory\n");
		err=-ENOMEM;
		goto init_no_ring_buffer;
		}
	
	printk("PI: enabling board\n");
	pi_stage_enable_card();
	return 0;
	/* gotos for error dropping - "the only place for gotos" - Linus */

	// dummy release ring buffer error here
init_no_ring_buffer:
	unregister_chrdev(pi_major,PI_MOD_NAME);
init_no_char_dev:
	iounmap(pi_stage.gpbus);
init_no_gpbus:
	release_mem_region(pci_resource_start(pi_stage.dev,PI_STAGE_GPBUS),PI_STAGE_GPBUS_SIZE);
init_gpbus_memory_error:
	iounmap(pi_stage.control);
init_no_control:
	release_mem_region(pci_resource_start(pi_stage.dev,PI_STAGE_HPICSR),PI_STAGE_HPICSR_SIZE);
init_hpicsr_memory_error:
	// nothing here
	
	
	return err;
	}


/************************************************************************/
/* CLEAN								*/
void cleanup_module(void){
int res;
	/* free alloc'd resources, set to NULL in case of reentrance	*/
	printk("PI: cleanup_module: Matt Clark oct 2003- device %d\n",pi_major);
	pi_ms_sleep(DEBUG_PAUSE);
	if(proc_file){
		printk("slm: cleanup: remove proc\n");
		pi_ms_sleep(DEBUG_PAUSE);
		remove_proc_entry(PI_MOD_NAME,NULL);
		}
	//free the memeory - we don't want to leak here
	kfree(pi_stage.events);
	release_mem_region(pci_resource_start(pi_stage.dev,PI_STAGE_GPBUS),PI_STAGE_GPBUS_SIZE);
	release_mem_region(pci_resource_start(pi_stage.dev,PI_STAGE_HPICSR),PI_STAGE_HPICSR_SIZE);
	res=unregister_chrdev(pi_major,PI_MOD_NAME);
	}


/************************************************************************/
/* IOCTL								*/
static int pi_ioctl(struct inode * inode, struct file *file, unsigned int cmd, unsigned long arg){
int i,length;
PI_DATA message[4];
	length=0;
	
	switch(cmd){
		case PI_STAGE_HARDRESET:
			/* resets the module count to 1 so that the 			*
			 * module can be removed if crashed				*/
			printk("PI: ioctl: HARDRESET: attempting mod_use=1\n");
						/* pull it to 0 			*/
			while(MOD_IN_USE)MOD_DEC_USE_COUNT;
			/* inc once for this open 					*/
			MOD_INC_USE_COUNT;
			/* set locked to FALSE 						*/
			pi_stage.locked=FALSE;
			printk("PI: Done\n");
			break;
		case PI_RESET_BOARD:
			pi_reset_board();
			break;	
		// all real activity goes on here
		// we will have the variety of gets and sets here
		// PI_SET_0 - just a command
		// PI_SET_1 - one 16 bit word for data
		// PI_SET_2 - two 16 bit word for data
		// PI_SET_3 - one arg (goes to data) then two 16 bit data
		//            of three 16 bit data
		case PI_SET_3:
			length++;
		case PI_SET_2:
			length++;
		case PI_SET_1:
			length++;
		case PI_SET_0:
			length++; // for the cmd
			
			copy_from_user((void *)message,(void *)arg,length*sizeof(PI_DATA));
			rdtscl(pi_prof1);
			spin_lock(pi_lock);
			pi_send_cmd(message[0]);
			for(i=1;i<length;i++){
				rdtscl(pi_prof2);
				pi_send_data(message[i]);
				}	
			spin_unlock(pi_lock);
			rdtscl(pi_prof3);
			return 0;	
			break;	

		// now simple reads
		case PI_GET_2:
			length++;
		case PI_GET_1:
			length++;
			length++; // for the cmd
			
			copy_from_user((void *)message,(void *)arg,sizeof(PI_DATA));
			rdtscl(pi_prof1);
			spin_lock(pi_lock);
			pi_send_cmd(message[0]);
			for(i=1;i<length;i++){
				rdtscl(pi_prof2);
				message[i]=(PI_DATA)pi_read_data();
				}	
			spin_unlock(pi_lock);
			rdtscl(pi_prof3);
			copy_to_user((void *)arg,(void *)message,sizeof(PI_DATA)*(length+1));
			return 0;
			break;	
			
		// now reads with arguments (rarely used)
		case PI_SET_GET_2:
			length++;
		case PI_SET_GET_1:
			length++;
			length++;  // for the cmd
			length++;  // for the argument

			copy_from_user((void *)message,(void *)arg,sizeof(PI_DATA));
			rdtscl(pi_prof1);
			spin_lock(pi_lock);
			pi_send_cmd(message[0]);
			pi_send_data(message[1]);
			for(i=2;i<length;i++){
				rdtscl(pi_prof2);
				message[i]=(PI_DATA)pi_read_data();
				}	
			spin_unlock(pi_lock);
			rdtscl(pi_prof3);
			copy_to_user((void *)arg,(void *)message,sizeof(PI_DATA)*length);
			return 0;
			break;
		default:
			/* during development report other events, after dev do nowt	*/	
			printk("PI_STAGE: rec'd unknown ioctl %ud\n",cmd);
		}
	return 0;
	}
// this reads the ring buffer as filled by pi_tasklet
// we will return 0 if there is no data pending and we
// are set to async data reading otherwise we will block
static ssize_t pi_read(struct file * file, char * buf, size_t count, loff_t *ppos){
int n=0;
	if(sizeof(struct pi_stage_event_t)>count)
		return -ERANGE;
	if(pi_stage.event_head==pi_stage.event_tail)
		return 0;
	spin_lock(pi_lock);
	while((pi_stage.event_tail!=pi_stage.event_head) && (n<(count-sizeof(struct pi_stage_event_t)))){
		copy_to_user(buf+n,pi_stage.events+pi_stage.event_tail,sizeof(struct pi_stage_event_t));
		pi_stage.event_tail++;
		if(pi_stage.event_tail>=PI_N_EVENTS)pi_stage.event_tail=0;
		n+=sizeof(struct pi_stage_event_t);
		}
	spin_unlock(pi_lock);
	return n;
	}

/****************************************************************************************/
/* INTERRUPT 										*/
/* interrupt handler									*/

static void pi_int_handler(int irq, void *dev_id,struct pt_regs *regs){
u32 event;

	pi_stage.ints_all++;
/* probe to see if this is ours								*/
	event=pi_read_control(PI_STAGE_INTEVENTSET);
	if(event & PI_STAGE_ALLINTS){
		// clear / acknowledge the interrupt
		/* 1st acknowledge the top level interrupt and clear it			*/
		pi_write_control(PI_STAGE_INTEVENTCLEAR,event & PI_STAGE_ALLINTS);
		/* then check for lower level interrupts GPIO3 and GPINT		*/
		/* if these are rec'd then we have an interesting event			*/
		if(event & PI_STAGE_INTGPIO3){
			pi_stage.ints_io++;
			}
		if(event & PI_STAGE_GPINT){
			pi_stage.ints_axis++;
			// run the tasklet to get this event
			tasklet_schedule(&pi_tasklet);
			}
		/* Since this is our interrupt keep track of the number 		*/
		pi_stage.ints_board++;
		}
	}

static void pi_stage_enable_card(void){
	// twang the hardware to get it going
	
	// write the correct sequence to the PCI command register
	// sets memory enable, parity error enable and sytem error enable (probably not required)
	pci_write_config_word(pi_stage.dev,PI_STAGE_PCIREG_CMD,PI_STAGE_PCIREG_CMD_OURS);
	
	// allow interrupts from GPIO3 to be routed
	pci_write_config_byte(pi_stage.dev,PI_STAGE_PCIREG_GPIO_SELECT,PI_STAGE_PCIREG_GPIO_GPIO3PIN);
	
	// make interrupt when GPIO3 is low (data ready?)
	pci_write_config_byte(pi_stage.dev,PI_STAGE_PCIREG_GPIO_INTTYPE,PI_STAGE_PCIREG_GPIO_LOW);
	
	// clear the interrupts
	pi_write_control(PI_STAGE_INTEVENTCLEAR,event & PI_STAGE_ALLINTS);
	
	// disable all the interrupts
	pi_write_control(PI_STAGE_INTMASKCLEAR,PI_STAGE_ALLINTS|PI_STAGE_ENABLEINT);

	// enable GPINT - this does not seem necessary on Athlon systems - this is a test setting
	// following line killed ritalin!
	// it is not required on ritalin and diesel!
	// anyway it shouldn't be here it should be with open....
	//pi_write_control(PI_STAGE_INTMASKSET,PI_STAGE_GPINT | PI_STAGE_ENABLEINT);
	//pi_write_control(PI_STAGE_INTMASKCLEAR,PI_STAGE_GPINT);
	}
/* 
// These aren't needed but left here for documentation
inline static void pi_enable_io_int(void){
	// line 754 of driver/Pci2040IRQ.cpp
	pi_write_control(PI_STAGE_INTMASKSET,PI_STAGE_INTGPIO3 | PI_STAGE_ENABLEINT);
	}
inline static void pi_disable_io_int(void){
	// line 513 of driver/Pci2040IRQ.cpp
	pi_write_control(PI_STAGE_INTMASKCLEAR,PI_STAGE_INTGPIO3);
	}
inline static void pi_enable_axis_int(void){
	// line 465 of driver/Pci2040IRQ.cpp
	pi_write_control(PI_STAGE_INTMASKSET,PI_STAGE_GPINT | PI_STAGE_ENABLEINT);
	}

inline static void pi_disable_axis_int(void){
	// line 130 of driver/Pci2040IRQ.cpp
	pi_write_control(PI_STAGE_INTMASKCLEAR,PI_STAGE_GPINT);
	}
*/
/****************************************************************************************/
/* /PROC filesystem									*/
/* implement a read only /proc filesystem that gives useful info about the card		*/
/* fill 1 page max- so ignore start, offset and we don't need data -no private data	*/

static int pi_proc_read(char * buf, char **start, off_t offset, int count, int *eof, void *data){
int w=0;
	/* force and eof for "normal library functions"	*/
	if(offset!=0){
		*eof=1;
		return 0;
		}
	if(count<70){
		/* not enough space for a warning*/
		return 0;
		}
	if((count>=70)&&(count<1024)){
		printk("Sorry: due to a lazy programmer the /proc buffer size must be > 1k\n");
		}
//	size=count<PAGE_SIZE?count:PAGE_SIZE;
/* now the idea for this /proc entry is to give basic pi_stage.card info for the user	*/
        w+=sprintf(buf+w,"PI_STAGE: PI PCI stage controller driver  %s (c)mc 2003\n",VERSION);  
	w+=sprintf(buf+w,"\nInterrupt status\n");
	w+=sprintf(buf+w,"Interrupts rec'd (all)   %d\n",pi_stage.ints_all);
	w+=sprintf(buf+w,"Interrupts rec'd (io)    %d\n",pi_stage.ints_io);
	w+=sprintf(buf+w,"Interrupts rec'd (axis)  %d\n",pi_stage.ints_axis);
	w+=sprintf(buf+w,"Interrupts rec'd (board) %d\n",pi_stage.ints_board);

	w+=sprintf(buf+w,"\nProfileing info\n");
	w+=sprintf(buf+w,"2-1         %d\n",(int)(pi_prof2-pi_prof1));
	w+=sprintf(buf+w,"3-2         %d\n",(int)(pi_prof3-pi_prof2));
	//w+=sprintf(buf+w,"4-3         %d\n",(int)(pi_prof4-pi_prof2));
	w+=sprintf(buf+w,"\nLast error status:\n");
	switch(pi_stage.error){
		case PI_NO_ERROR:
			w+=sprintf(buf+w,"No errors\n");
			break;
		case PI_LOOPED_OUT:
			w+=sprintf(buf+w,"Warning: Looped out - report to driver author\n");
			break;
		case PI_LOOPED_OUT_FATAL:
			w+=sprintf(buf+w,"Fatal: Looped out - report to driver author\n");
			w+=sprintf(buf+w,"Fatal: You may need to reboot - sorry\n");
			break;
		case PI_RESET_FAIL:
			w+=sprintf(buf+w,"Error: Reset failed - probably ok but please report to driver author\n");
			break;
		default:
			w+=sprintf(buf+w,"Undefined error\n");
		}		
	
	w+=sprintf(buf+w,"\nRing buffer status\n");
	w+=sprintf(buf+w,"Async events rec'd      %d\n",pi_stage.event_index);
	w+=sprintf(buf+w,"Event head              %d\n",pi_stage.event_head);
	w+=sprintf(buf+w,"Event tail              %d\n",pi_stage.event_tail);
        if(pi_stage.async==TRUE){
		w+=sprintf(buf+w,"Asynchronous notification enabled\n");  
		}
        w+=sprintf(buf+w,"Null proc entry %d\n\n",w);  
	return w;
	}	


/****************************************************************************************/
/* OPEN											*/
/* in this case- global device structure init'd at driver install this doen't do much	*/
/* simply enforce single open access */
/* global locking here prevents ioctl */
/* if multiple cards are needed- use minor to reference their structures		*/

static int pi_open(struct inode *inode, struct file *file){
int err;
/* allow double open during debugging -matt */
     if(pi_stage.locked==TRUE) return -EBUSY;
	err=request_irq(pi_stage.interrupt,pi_int_handler,SA_SHIRQ,PI_IRQ_ID,(void *)&pi_stage);
	if(err!=0){
		printk("PI: open: Could not register interrupt %d, returning error\n",pi_stage.interrupt);
		return err;
		}
	printk("PI: open: registered interrupt %d Ok\n",pi_stage.interrupt);
        pi_stage.locked=TRUE;
        MOD_INC_USE_COUNT;
        //printk("PI:open:done\n");
	//
	//attempt to get ints on "hong"
	pi_write_control(PI_STAGE_INTMASKSET,PI_STAGE_GPINT | PI_STAGE_ENABLEINT);
	pi_write_control(PI_STAGE_INTMASKCLEAR,PI_STAGE_GPINT);
        return 0;
        }

/****************************************************************************************/
/* RELEASE										*/
/* As open hardly does anything this does little as well				*/

static int pi_release(struct inode *inode, struct file *file) {
	
        MOD_DEC_USE_COUNT;
/* release the interrupt */
	// added for "hong"
	pi_write_control(PI_STAGE_INTMASKCLEAR,PI_STAGE_ALLINTS|PI_STAGE_ENABLEINT);

	free_irq(pi_stage.interrupt,(void *)&pi_stage);		
/* release the device */
        pi_stage.locked=FALSE;
/* release any async process notification - not convinced this is necessary */
	pi_async(-1,file,0);
        return 0;
        }
/* PI_STAGE utlities here */

/* delay by going to sleep- note that the smallest delay is 1000/HZ ms	(10ms typically)*/

static void pi_ms_sleep(int ms){
int jif;
	jif=1+ms*HZ/1000; 	/* round up 	*/
	if(jif<1)jif=1;		/* not required	iff ms always +ve */
	current->state=TASK_INTERRUPTIBLE;
	schedule_timeout(jif);
	}


void pi_write_control_w(int place,u16 word){
	writew(word,pi_stage.control+place);
	}

void pi_write_control(int place,u32 word){
	 writel(word,pi_stage.control+place);
	}

inline u32 pi_read_control(int place){
	return  readl(pi_stage.control+place);
	}

inline u16 pi_read_control_w(int place){
	return  readw(pi_stage.control+place);
	}

inline int pi_test_io_ok(void){
int err;
char val;
	// test the io condition by polling the config space
	// - remember I didn't design it! -matt
	// based on line: 210 of driver/Pci2040IRQ.cpp
	err=pci_read_config_byte(pi_stage.dev,PI_STAGE_PCIREG_GPIO_INDATA,&val);
	if(err!=0){
		printk("PI: test_io: pci read config error\n");
		return FALSE;
		}
	if((val & 0x08)==0)
		return FALSE;
	return TRUE;
	}

// nicely fixed up now
// BUT must be used in the right place or else
// basically before any reads/writes
// error conditions to remain during alpha/beta testing
inline void pi_busy_wait_io(void){
	int i=0;
	// Wait for a short time
	while((i<PI_MAX_WAIT) & (pi_test_io_ok()==FALSE))
		i++;
	// Not enough? well then wait a long time in a system kindly way
	if(i==PI_MAX_WAIT){
		printk("PI: Warning: Driver error: looped out - please contact driver author\n");
		pi_stage.error=PI_LOOPED_OUT;
		while((i<PI_MAX_WAIT*2) & (pi_test_io_ok()==FALSE)){
			pi_ms_sleep(1);
			i++;
			}
		if(i==(PI_MAX_WAIT*2)){
			pi_stage.error=PI_LOOPED_OUT_FATAL;
			printk("PI: Driver error: fatal: you must reboot to recover - contact driver author\n");
			}
		}
	}


inline void pi_send_cmd(u16 command){
	pi_busy_wait_io();
	writew(command,pi_stage.gpbus+PI_STAGE_GP_CMD);
	}

inline void pi_send_data(u16 data){
	pi_busy_wait_io();
	writew(data,pi_stage.gpbus+PI_STAGE_GP_DATA);
	}

inline u16 pi_read_data(void){
	pi_busy_wait_io();
	return readw(pi_stage.gpbus+PI_STAGE_GP_DATA);
	}

void pi_reset_board(void){
int err;
	// reset the board - don't poll just wait here it is more convenient
	// the correct sequence is 
	// reset
	// read PI_GET_HOST_IO_ERROR
	// read PI_GET_EVENT_STATUS
	// reset PI_RESET_EVENT_STATUS
	// via the PMD FAQ 
	// otherwise you can see spurious errors / events
	//   
	// sleeps here are determined by experiment
	// shorter sleeps can result in spurious errors
	// 
	// pi_send_cmd(PI_BOARD_RESET);
	// the reset can take a long time.
	// explicitly use writew here rather than pi_send_cmd
	// to avoid spurious LOOPED_OUT error
	writew(PI_BOARD_RESET,pi_stage.gpbus+PI_STAGE_GP_CMD);
	pi_ms_sleep(10);
	printk("PI: reset performed\n");
	pi_send_cmd(0x00);
	pi_ms_sleep(10);
	pi_send_cmd(PI_GET_HOST_IO_ERROR);
	pi_ms_sleep(10);
	err=pi_read_data();
	if((err!=0x00) & (err != 0x01)){
		printk("PI: PI_GET_HOST_IO_ERROR returned 0x%x (Should be 0x00 or 0x01)\n",(int)err);
		pi_stage.error=PI_RESET_FAIL;
		}
	pi_send_cmd(PI_GET_EVENT_STATUS);
	err=pi_read_data();
	pi_send_cmd(PI_RESET_EVENT_STATUS);
	pi_send_data(0);
	// reset the ring buffer
	pi_stage.event_head=0;
	pi_stage.event_tail=0;
	}

void pi_getasyncevent(unsigned long temp){
u32 axis,event,i;
unsigned long high,low;
	spin_lock(pi_lock);
	// safe in here
	// get the axis which interrupted
	pi_send_cmd(PI_GET_INTERRUPT_AXIS);
	axis=pi_read_data();
	pi_stage.event_index++;
	for(i=0;i<pi_stage.naxis;i++){
		event=0;
		if(axis & (1<<i)){
			//get event
			pi_send_cmd(PI_GET_EVENT_STATUS | (i<<8));
			event=pi_read_data();
			rdtsc(low,high);
			// we have all we need, save and quit
			pi_stage.events[pi_stage.event_head].index	=pi_stage.event_index;
			pi_stage.events[pi_stage.event_head].axis	=i;
			pi_stage.events[pi_stage.event_head].ts_high	=high;
			pi_stage.events[pi_stage.event_head].ts_low	=low;
			pi_stage.events[pi_stage.event_head].event	=event;
			
			// inc the head to the next place and take care of collisions
			pi_stage.event_head++;
			if(pi_stage.event_head>=PI_N_EVENTS){
				pi_stage.event_head=0;
				}
			if(pi_stage.event_head==pi_stage.event_tail){
				// head and tail collided
				pi_stage.event_tail++;
				if(pi_stage.event_tail>=PI_N_EVENTS){
					pi_stage.event_tail=0;
					}
				// this should be commented out as we don't want to flood
				// the logs with spurious messages.... - matt July 2003
				// Keep for now for debugging
				printk("PI: ring buffer collision\n");
				}
			}
			// reset event
		pi_send_cmd(PI_RESET_EVENT_STATUS | (i<<8));
		pi_send_data(0);
		}
	pi_send_cmd(PI_CLEAR_INTERRUPT);
	spin_unlock(pi_lock);
	if(pi_stage.async_queue!=NULL){
		kill_fasync(&(pi_stage.async_queue),SIGIO,POLL_IN);
		}
	}
	
/* kernel license GPL */
MODULE_AUTHOR("Matt Clark");
MODULE_DESCRIPTION("PI PCI card driver");
MODULE_LICENSE("GPL");
/* we don't export symbols */
EXPORT_NO_SYMBOLS;
